//
//  MyScene.m
//  Flappy Felipe
//
//  Created by Main Account on 2/13/14.
//  Copyright (c) 2014 Razeware LLC. All rights reserved.
//

#import "MyScene.h"

typedef NS_ENUM(int, Layer) {
  LayerBackground,
  LayerObstacle,
  LayerForeground,
  LayerPlayer
};

typedef NS_OPTIONS(int, EntityCategory) {
  EntityCategoryPlayer = 1 << 0,
  EntityCategoryObstacle = 1 << 1,
  EntityCategoryGround = 1 << 2
};

// Gameplay - bird movement
static const float kGravity = -1500.0;
static const float kImpulse = 400.0;

// Gameplay - ground speed
static const float kGroundSpeed = 150.0f;

// Gameplay - obstacles positioning
static const float kGapMultiplier = 3.5;
static const float kBottomObstacleMinFraction = 0.1;
static const float kBottomObstacleMaxFraction = 0.6;

// Gameplay - obstacles timing
static const float kFirstSpawnDelay = 1.75;
static const float kEverySpawnDelay = 1.5;

// Looks
static const int kNumForegrounds = 2;

@interface MyScene() <SKPhysicsContactDelegate>
@end

@implementation MyScene {

  SKNode *_worldNode;
  
  float _playableStart;
  float _playableHeight;

  NSTimeInterval _lastUpdateTime;
  NSTimeInterval _dt;

  SKSpriteNode *_player;
  SKSpriteNode *_sombrero;

  CGPoint _playerVelocity;
  
  SKAction * _dingAction;
  SKAction * _flapAction;
  SKAction * _whackAction;
  SKAction * _fallingAction;
  SKAction * _hitGroundAction;
  SKAction * _popAction;
  SKAction * _coinAction;
  
  BOOL _hitGround;
  BOOL _hitObstacle;
  
  GameState _gameState;
  
}

-(id)initWithSize:(CGSize)size {
  if (self = [super initWithSize:size]) {
    
    _worldNode = [SKNode node];
    [self addChild:_worldNode];
    
    [self setupBackground];
    [self setupForeground];
    [self setupPlayer];
    [self setupSounds];
    [self startSpawning];
    [self setupSombrero];
    
    self.physicsWorld.contactDelegate = self;
    self.physicsWorld.gravity = CGVectorMake(0, 0);
    [self flapPlayer];
    
    _gameState = GameStatePlay;
    
  }
  return self;
}

- (void)setupSombrero {

  _sombrero = [SKSpriteNode spriteNodeWithImageNamed:@"Sombrero"];
  _sombrero.position = CGPointMake(31 - _sombrero.size.width/2, 29 - _sombrero.size.height/2);
  [_player addChild:_sombrero];
  

}

#pragma mark - Setup methods

- (void)setupBackground {
  SKSpriteNode *background = [SKSpriteNode spriteNodeWithImageNamed:@"Background"];
  background.anchorPoint = CGPointMake(0.5, 1);
  background.position = CGPointMake(self.size.width/2, self.size.height);
  background.zPosition = LayerBackground;
  [_worldNode addChild:background];
  
  _playableStart = self.size.height - background.size.height;
  _playableHeight = background.size.height;

  // 1
  CGPoint lowerLeft = CGPointMake(0, _playableStart);
  CGPoint lowerRight = CGPointMake(self.size.width, _playableStart);

  self.physicsBody = [SKPhysicsBody bodyWithEdgeFromPoint:lowerLeft toPoint:lowerRight];
  [self skt_attachDebugLineFromPoint:lowerLeft toPoint:lowerRight color:[UIColor redColor]];
  
  self.physicsBody.categoryBitMask = EntityCategoryGround;
  self.physicsBody.collisionBitMask = 0;
  self.physicsBody.contactTestBitMask = EntityCategoryPlayer;

}

- (void)setupForeground {
  for (int i = 0; i < kNumForegrounds; ++i) {
    SKSpriteNode *foreground = [SKSpriteNode spriteNodeWithImageNamed:@"Ground"];
    foreground.anchorPoint = CGPointMake(0, 1);
    foreground.position = CGPointMake(i * self.size.width, _playableStart);
    foreground.zPosition = LayerForeground;
    foreground.name = @"Foreground";
    [_worldNode addChild:foreground];
  }
}

- (void)setupPlayer {

  _player = [SKSpriteNode spriteNodeWithImageNamed:@"Bird0"];
  _player.position = CGPointMake(self.size.width * 0.2, _playableHeight * 0.4 + _playableStart);
  _player.zPosition = LayerPlayer;
  [_worldNode addChild:_player];
  
  CGFloat offsetX = _player.frame.size.width * _player.anchorPoint.x;
  CGFloat offsetY = _player.frame.size.height * _player.anchorPoint.y;
  CGMutablePathRef path = CGPathCreateMutable();
  CGPathMoveToPoint(path, NULL, 8 - offsetX, 4 - offsetY);
  CGPathAddLineToPoint(path, NULL, 8 - offsetX, 15 - offsetY);
  CGPathAddLineToPoint(path, NULL, 23 - offsetX, 29 - offsetY);
  CGPathAddLineToPoint(path, NULL, 39 - offsetX, 25 - offsetY);
  CGPathAddLineToPoint(path, NULL, 37 - offsetX, 7 - offsetY);
  CGPathAddLineToPoint(path, NULL, 24 - offsetX, 4 - offsetY);
  CGPathCloseSubpath(path);
  _player.physicsBody = [SKPhysicsBody bodyWithPolygonFromPath:path];
  
  [_player skt_attachDebugFrameFromPath:path color:[SKColor redColor]];

  _player.physicsBody.categoryBitMask = EntityCategoryPlayer;
  _player.physicsBody.collisionBitMask = 0;
  _player.physicsBody.contactTestBitMask = EntityCategoryObstacle | EntityCategoryGround;
  
}

- (void)setupSounds {
  _dingAction = [SKAction playSoundFileNamed:@"ding.wav" waitForCompletion:NO];
  _flapAction = [SKAction playSoundFileNamed:@"flapping.wav" waitForCompletion:NO];
  _whackAction = [SKAction playSoundFileNamed:@"whack.wav" waitForCompletion:NO];
  _fallingAction = [SKAction playSoundFileNamed:@"falling.wav" waitForCompletion:NO];
  _hitGroundAction = [SKAction playSoundFileNamed:@"hitGround.wav" waitForCompletion:NO];
  _popAction = [SKAction playSoundFileNamed:@"pop.wav" waitForCompletion:NO];
  _coinAction = [SKAction playSoundFileNamed:@"coin.wav" waitForCompletion:NO];
}

#pragma mark - Gameplay

- (SKSpriteNode *)createObstacle {
  SKSpriteNode *sprite = [SKSpriteNode spriteNodeWithImageNamed:@"Cactus"];
  sprite.zPosition = LayerObstacle;
  
  CGFloat offsetX = sprite.frame.size.width * sprite.anchorPoint.x;
  CGFloat offsetY = sprite.frame.size.height * sprite.anchorPoint.y;
  
  CGMutablePathRef path = CGPathCreateMutable();

  CGPathMoveToPoint(path, NULL, 16 - offsetX, 315 - offsetY);
  CGPathAddLineToPoint(path, NULL, 40 - offsetX, 314 - offsetY);
  CGPathAddLineToPoint(path, NULL, 51 - offsetX, 307 - offsetY);
  CGPathAddLineToPoint(path, NULL, 48 - offsetX, 0 - offsetY);
  CGPathAddLineToPoint(path, NULL, 3 - offsetX, 1 - offsetY);
  CGPathAddLineToPoint(path, NULL, 4 - offsetX, 128 - offsetY);
  CGPathAddLineToPoint(path, NULL, 3 - offsetX, 247 - offsetY);
  CGPathAddLineToPoint(path, NULL, 4 - offsetX, 308 - offsetY);

  CGPathCloseSubpath(path);

  sprite.physicsBody = [SKPhysicsBody bodyWithPolygonFromPath:path];
  
  [sprite skt_attachDebugFrameFromPath:path color:[SKColor redColor]];

  sprite.physicsBody.categoryBitMask = EntityCategoryObstacle;
  sprite.physicsBody.collisionBitMask = 0;
  sprite.physicsBody.contactTestBitMask = EntityCategoryPlayer;
  
  return sprite;
}

- (void)spawnObstacle {

  SKSpriteNode *bottomObstacle = [self createObstacle];
  bottomObstacle.name = @"BottomObstacle";
  float startX = self.size.width + bottomObstacle.size.width/2;
  
  float bottomObstacleMin = (_playableStart - bottomObstacle.size.height/2) + _playableHeight * kBottomObstacleMinFraction;
  float bottomObstacleMax = (_playableStart - bottomObstacle.size.height/2) + _playableHeight * kBottomObstacleMaxFraction;
  bottomObstacle.position = CGPointMake(startX, RandomFloatRange(bottomObstacleMin, bottomObstacleMax));
  [_worldNode addChild:bottomObstacle];
  
  SKSpriteNode *topObstacle = [self createObstacle];
  topObstacle.name = @"TopObstacle";
  topObstacle.zRotation = DegreesToRadians(180);
  topObstacle.position = CGPointMake(startX, bottomObstacle.position.y + bottomObstacle.size.height/2 + topObstacle.size.height/2 + _player.size.height * kGapMultiplier);
  [_worldNode addChild:topObstacle];

  float moveX = self.size.width + topObstacle.size.width;
  float moveDuration = moveX / kGroundSpeed;
  SKAction *sequence = [SKAction sequence:@[
    [SKAction moveByX:-moveX y:0 duration:moveDuration],
    [SKAction removeFromParent]
  ]];
  
  [topObstacle runAction:sequence];
  [bottomObstacle runAction:sequence];


}

- (void)startSpawning {

  SKAction *firstDelay = [SKAction waitForDuration:kFirstSpawnDelay];
  SKAction *spawn = [SKAction performSelector:@selector(spawnObstacle) onTarget:self];
  SKAction *everyDelay = [SKAction waitForDuration:kEverySpawnDelay];
  SKAction *spawnSequence = [SKAction sequence:@[spawn, everyDelay]];
  SKAction *foreverSpawn = [SKAction repeatActionForever:spawnSequence];
  SKAction *overallSequence = [SKAction sequence:@[firstDelay, foreverSpawn]];
  [self runAction:overallSequence withKey:@"Spawn"];

}

- (void)stopSpawning {
  [self removeActionForKey:@"Spawn"];
  [_worldNode enumerateChildNodesWithName:@"TopObstacle" usingBlock:^(SKNode *node, BOOL *stop) {
    [node removeAllActions];
  }];
  [_worldNode enumerateChildNodesWithName:@"BottomObstacle" usingBlock:^(SKNode *node, BOOL *stop) {
      [node removeAllActions];
    }];
}

- (void)flapPlayer {

  // Play sound
  [self runAction:_flapAction];

  // Apply impulse
  _playerVelocity = CGPointMake(0, kImpulse);
  
  // Move sombrero
  SKAction *moveUp = [SKAction moveByX:0 y:12 duration:0.15];
  moveUp.timingMode = SKActionTimingEaseInEaseOut;
  SKAction *moveDown = [moveUp reversedAction];
  [_sombrero runAction:[SKAction sequence:@[moveUp, moveDown]]];
  
}

-(void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event {
  switch (_gameState) {
  case GameStateMainMenu:
    break;
  case GameStateTutorial:
    break;
  case GameStatePlay:
    [self flapPlayer];
    break;
  case GameStateFalling:
    break;
  case GameStateShowingScore:
    [self switchToNewGame];
    break;
  case GameStateGameOver:
    break;
  }
}

#pragma mark - Switch state

- (void)switchToFalling {
  _gameState = GameStateFalling;
  
  // Transition code...
  [self runAction:[SKAction sequence:@[
    _whackAction,
    [SKAction waitForDuration:0.1],
    _fallingAction]]];
  
  [_player removeAllActions];
  [self stopSpawning];
}

- (void)switchToShowScore {

  _gameState = GameStateShowingScore;
 
  [_player removeAllActions];
  [self stopSpawning];
  
}

- (void)switchToNewGame {

  [self runAction:_popAction];

  SKScene *newScene = [[MyScene alloc] initWithSize:self.size];
  SKTransition *transition = [SKTransition fadeWithColor:[SKColor blackColor] duration:0.5];
  [self.view presentScene:newScene transition:transition];
}

#pragma mark - Updates

- (void)checkHitGround {
  if (_hitGround) {
    _hitGround = NO;
    _playerVelocity = CGPointZero;
    _player.position = CGPointMake(_player.position.x, _playableStart + _player.size.width/2);
    _player.zRotation = DegreesToRadians(-90);
    [self runAction:_hitGroundAction];
    [self switchToShowScore];
  }
}

- (void)checkHitObstacle {
  if (_hitObstacle) {
    _hitObstacle = NO;
    [self switchToFalling];
  }
}

- (void)updatePlayer {
  
  // Apply gravity
  CGPoint gravity = CGPointMake(0, kGravity);
  CGPoint gravityStep = CGPointMultiplyScalar(gravity, _dt);
  _playerVelocity = CGPointAdd(_playerVelocity, gravityStep);

  // Apply velocity
  CGPoint velocityStep = CGPointMultiplyScalar(_playerVelocity, _dt);
  _player.position = CGPointAdd(_player.position, velocityStep);
  
  // Temporary halt when hits ground
//  if (_player.position.y - _player.size.height/2 <= _playableStart) {
//    _player.position = CGPointMake(_player.position.x, _playableStart + _player.size.height/2);
//    return;
//  }
  
}

- (void)updateForeground {

  [_worldNode enumerateChildNodesWithName:@"Foreground" usingBlock:^(SKNode *node, BOOL *stop) {
    SKSpriteNode *foreground = (SKSpriteNode *)node;
    CGPoint moveAmt = CGPointMake(-kGroundSpeed * _dt, 0);
    foreground.position = CGPointAdd(foreground.position, moveAmt);
    
    if (foreground.position.x < -foreground.size.width) {
      foreground.position = CGPointAdd(foreground.position, CGPointMake(foreground.size.width * kNumForegrounds, 0));
    }
    
  }];

}

-(void)update:(CFTimeInterval)currentTime {
  if (_lastUpdateTime) {
    _dt = currentTime - _lastUpdateTime;
  } else {
    _dt = 0;
  }
  _lastUpdateTime = currentTime;

  switch (_gameState) {
  case GameStateMainMenu:
    break;
  case GameStateTutorial:
    break;
  case GameStatePlay:
    [self checkHitGround];
    [self checkHitObstacle];
    [self updateForeground];
    [self updatePlayer];
    break;
  case GameStateFalling:
    [self checkHitGround];
    [self updatePlayer];
    break;
  case GameStateShowingScore:
    break;
  case GameStateGameOver:
    break;
  }
}

#pragma mark - Collision Detection

- (void)didBeginContact:(SKPhysicsContact *)contact {
  SKPhysicsBody *other = (contact.bodyA.categoryBitMask == EntityCategoryPlayer ? contact.bodyB : contact.bodyA);
  if (other.categoryBitMask == EntityCategoryGround) {
    _hitGround = YES;
    return;
  }
  if (other.categoryBitMask == EntityCategoryObstacle) {
    _hitObstacle = YES;
    return;
  }
}

@end
